# -----------------------------------------------------------------------------
# Endroid - XMPP Bot - PeriodicPinger plugin
# Copyright 2014, Ensoft Ltd.
# Created by Patrick Stevens
# -----------------------------------------------------------------------------

"""
Periodically XMPP-pings each XMPP server Endroid is communicating with.
It is hoped that this will alleviate problems related to servers dropping our
connections.
"""


import logging
import itertools

import twisted.internet.task

from endroid.pluginmanager import Plugin


class PeriodicPinger(Plugin):
    """
    Wrapper to a Cron job which periodically pings a user on each server.
    This plugin has no XMPP-interface by which the user may interact with it.
    """

    # Class variable to synchronise plugin so that even if we 
    # instantiate more than once, we still only have one Cron job
    task = None

    # Add the hidden attribute so that this plugin doesn't appear in help
    hidden = True

    def endroid_init(self):
        cls = type(self)
        if cls.task is None:
            # Get the set of servers that endroid could communicate to. This
            # is static so we do this once at init time.
            self.servers = self._get_servers()
            self.log = logging.getLogger(__name__)
            self.interval = self.vars.get('interval', 10)
            cls.task = twisted.internet.task.LoopingCall(self._ping_servers)
            cls.task.start(self.interval)
            self.log.info("Periodic pinger set up for every {} seconds"
                          .format(self.interval))

    def _get_servers(self):
        """
        Get the set of servers that endroid could communicate to (using the
        list of registered users and rooms)
        """
        servers = set()
        for user in self.usermanagement.users():
            servers.add(self.usermanagement.get_host(user))
        for room in self.usermanagement.rooms():
            servers.add(self.usermanagement.get_host(room))

        return servers

    def _ping_servers(self):
        """Pings each server Endroid knows about."""
        self.log.debug('PeriodicPinger: Pinging {}...'.format(self.servers))

        for server in self.servers:
            def _fail(fail, server):
                # If the server returns an error then we don't really care, 
                # since all we want to do is ensure server-to-server 
                # connections are kept open by sending ping traffic and we have 
                # achieved this. So just log for debugging purposes.
                logging.debug("Failed to ping {}: {}"
                              .format(server, str(fail.value)))
            def _success(response):
                pass
            d = self.usermanagement.ping(server)
            d.addCallback(_success)
            d.addErrback(_fail, server)


